# Copyright (c) 2023, Idiap Research Institute. All rights reserved.
#
# This work is made available under CC BY 4.0 license.
#
# Written by:
# Eklavya Sarkar <eklavya.sarkar@idiap.ch>

import os

import librosa
import numpy as np
import pandas as pd
import torch
import torchaudio
from torch.utils.data import Dataset


class InfantMarmosetsVox(Dataset):
    """
    This dataloader reads the InfantMarmosetsVox dataset
    data, and constructs a PyTorch Dataset.

    From the authors:
    Sarkar et al.: https://arxiv.org/abs/2305.14035
    Zhang et al.: https://asa.scitation.org/doi/pdf/10.1121/1.5047743
    """

    def __init__(
        self,
        target_sr=None,
        selected_labels=None,
    ):
        self.target_sr = target_sr
        self.selected_labels = selected_labels
        self.filelist = self._construct_dataframe()

    def __len__(self):
        """Returns the length of the dataset."""
        return len(self.filelist)

    def __getitem__(self, index):
        """Reads and processes an audio signal and
        returns a tuple with its corresponding label.

        Parameters
        ----------
        index : int
            Index of the data sample.

        Returns
        -------
        tuple
            The audio signal and its chosen label.
        """
        # Get variables
        audio_sample_path = self._get_audio_sample_path(index)
        signal, sr = self._load_audio_segment(index, audio_sample_path)

        # Preprocess (if necessary)
        signal = self._resample_if_necessary(signal, sr)  # Downsample
        signal = self._mix_down_if_necessary(signal)  # Mono-channel

        # Get labels
        calltype_label = self._get_audio_sample_calltype_label(index)
        caller_label = self._get_audio_sample_caller_label(index)

        # Label to return based on selection
        labels = {
            "calltype": calltype_label,
            "caller": caller_label,
        }
        return_label = labels[self.selected_labels]

        # Return
        return signal, return_label

    def _construct_dataframe(self):
        """Constructs and returns a pd.DataFrame object
        based on the annotations in the `labels.csv` file.

        Returns
        -------
        pd.DataFrame
            A pandas DataFrame containing all the vocalization
            segments and their corresponding calltype and caller labels.
        """
        # Read vocalization labels
        df = pd.read_csv("labels.csv")

        df["twinID"] = df.filename.apply(lambda x: x.split("_")[1][-1])
        df["individualID"] = df.filename.apply(lambda x: x.split("_")[2][-1])
        df["date"] = df.filename.apply(lambda x: x.split("_")[0])

        # Process
        df.drop(columns=["filename"], inplace=True)
        df.rename(columns={"calltype": "calltypeID"}, inplace=True)
        df.rename(columns={"caller": "callerID"}, inplace=True)

        # Get paths
        df["audio_file"] = df.apply(
            lambda row: f"data/twin_{row.twinID}/{row.date}_Twin{row.twinID}_marmoset{row.individualID}.wav",
            axis=1,
        )

        # Check all files exist
        df["exists"] = df["audio_file"].astype(str).map(os.path.exists)

        # Drop rows where the audio files don't exist
        df = df[df.exists == True]

        # Drop rows where CalltypeID is '11' or '12' (silence or noise)
        df = df[df.calltypeID != 11]
        df = df[df.calltypeID != 12]

        # Reorder dataframe
        ordered_cols = [
            "audio_file",
            "start",
            "end",
            "duration",
            "calltypeID",
            "callerID",
        ]

        # Keep only relevant columns
        df = df[ordered_cols]

        # Reset index to account for dropped rows
        df.reset_index(drop=True, inplace=True)

        # Vocalization column
        df["vocID"] = df.index

        # Return
        return df

    def _load_audio_segment(self, index, audio_sample_path):
        """Reads and normalizes an audio vocalization segment.

        Parameters
        ----------
        index : int
            Index of dataframe.
        audio_sample_path : str
            Path to the audio file selected by `index`.

        Returns
        -------
        tuple
            A tuple of the normalized signal and native sampling rate.
        """
        start = self.filelist.start.iloc[index]
        end = self.filelist.end.iloc[index]
        signal, sr = librosa.load(
            audio_sample_path,
            sr=None,
            mono=True,
            # dtype="float32", # default
            offset=start,
            duration=end - start,
        )

        # Channels first like with torchaudio.load
        signal /= np.max(np.abs(signal))
        signal = np.expand_dims(signal, axis=0)
        signal = torch.tensor(signal)
        return signal, sr

    def _resample_if_necessary(self, signal, sr):
        """Downsamples the provided signal to the defined `target_sr`.

        Parameters
        ----------
        signal : torch.tensor
            A tensor containing the audio signal.
        sr : int
            Native sampling rate of `signal`.

        Returns
        -------
        torch.tensor
            Returns the downsampled signal.
        """
        if sr != self.target_sr:
            resampler = torchaudio.transforms.Resample(sr, self.target_sr)
            signal = resampler(signal)
        return signal

    def _mix_down_if_necessary(self, signal):
        """Converts signal to mono.channel if it has more than 1 channel.

        Parameters
        ----------
        signal : torch.tensor
            The audio signal.

        Returns
        -------
        torch.tensor
            The mono-channel signal.
        """
        if signal.shape[0] > 1:  # Channels at index 0
            signal = torch.mean(signal, dim=0, keepdim=True)
        return signal

    def _get_audio_sample_path(self, index):
        """Returns the audio path of a selected signal."""
        return self.filelist.audio_file.iloc[index]

    def _get_audio_sample_calltype_label(self, index):
        """Returns the calltype label of a selected signal."""
        return self.filelist.calltypeID.iloc[index]

    def _get_audio_sample_caller_label(self, index):
        """Returns the caller label of a selected signal."""
        return self.filelist.callerID.iloc[index]


if __name__ == "__main__":
    _ = InfantMarmosetsVox()
