# Copyright (c) 2023, Idiap Research Institute. All rights reserved.
#
# This work is made available under CC BY 4.0 license.
#
# Written by:
# Eklavya Sarkar <eklavya.sarkar@idiap.ch>

import torch
from torch.utils.data import DataLoader

from infantmarmosetsvox import InfantMarmosetsVox


def pad_sequence(batch):
    # Make all tensor in a batch the same length by padding with zeros
    batch = [item.t() for item in batch]
    batch = torch.nn.utils.rnn.pad_sequence(batch, batch_first=True, padding_value=0.0)
    return batch.permute(0, 2, 1)


def collate_fn(batch):
    # Iterate
    tensors, targets = [], []
    for waveform, label in batch:
        tensors += [waveform]
        targets += [torch.tensor(label)]

    # Group the list of tensors into a batched tensor
    tensors = pad_sequence(tensors)
    targets = torch.stack(targets)

    # Return
    return tensors, targets


def main():
    # Device
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    print("Device:", device)

    # Variables
    LABELS_SELECTION = ("calltype", "caller")
    TARGET_SAMPLE_RATE = 16000

    # Instantiate Dataset
    data = InfantMarmosetsVox(
        target_sr=TARGET_SAMPLE_RATE,
        selected_labels=LABELS_SELECTION[0],
    )

    # Instantiate Dataloader
    dataloader = DataLoader(data, batch_size=16, collate_fn=collate_fn)
    print(f"There are {len(data)} data points and labels in the dataset.")

    # Iterate
    x, y = next(iter(dataloader))
    print("Sample data point and label shapes from a single batch (size=16):")
    print("x shape:", x.shape)
    print("y shape:", y.shape)


if __name__ == "__main__":
    main()
